'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createElement = createElement;
exports.getPixelRatio = getPixelRatio;
exports.createNodeRescalingFunction = createNodeRescalingFunction;
/**
 * Sigma.js Rendering Utils
 * ===========================
 *
 * Helpers used by most renderers.
 */

/**
 * Function used to create DOM elements easily.
 *
 * @param  {string} tag        - Tag name of the element to create.
 * @param  {object} attributes - Attributes map.
 * @return {HTMLElement}
 */
function createElement(tag, attributes) {
  var element = document.createElement(tag);

  if (!attributes) return element;

  for (var k in attributes) {
    if (k === 'style') {
      for (var s in attributes[k]) {
        element.style[s] = attributes[k][s];
      }
    } else {
      element.setAttribute(k, attributes[k]);
    }
  }

  return element;
}

/**
 * Function returning the browser's pixel ratio.
 *
 * @return {number}
 */
function getPixelRatio() {
  var screen = window.screen;

  if (typeof screen.deviceXDPI !== 'undefined' && typeof screen.logicalXDPI !== 'undefined' && screen.deviceXDPI > screen.logicalXDPI) return screen.systemXDPI / screen.logicalXDPI;else if (typeof window.devicePixelRatio !== 'undefined') return window.devicePixelRatio;

  return 1;
}

/**
 * Default rescale options.
 */
var DEFAULT_NODE_RESCALE_OPTIONS = {
  mode: 'inside',
  margin: 0,
  minSize: 1,
  maxSize: 8
};

var DEFAULT_EDGE_RESCALE_OPTIONS = {
  minSize: 0.5,
  maxSize: 1
};

// TODO: should we let the user handle size through, for instance, d3 scales?
// TODO: should we put the rescaling in the camera itself?

/**
 * Factory returning a function rescaling the given node's position and/or size.
 *
 * @param  {object}   options - Options.
 * @param  {object}   extent  - Extent of the graph.
 * @return {function}
 */
function createNodeRescalingFunction(options, extent) {
  options = options || {};

  var mode = options.mode || DEFAULT_NODE_RESCALE_OPTIONS.mode,
      height = options.height || 1,
      width = options.width || 1;

  var maxX = extent.maxX,
      maxY = extent.maxY,
      minX = extent.minX,
      minY = extent.minY;


  var hx = (maxX + minX) / 2,
      hy = (maxY + minY) / 2;

  var scale = mode === 'outside' ? Math.max(width / Math.max(maxX - minX, 1), height / Math.max(maxY - minY, 1)) : Math.min(width / Math.max(maxX - minX, 1), height / Math.max(maxY - minY, 1));

  var fn = function fn(data) {
    return {
      x: (data.x - hx) * scale,
      y: (data.y - hy) * scale
    };
  };

  fn.inverse = function (data) {
    return {
      x: data.x / scale + hx,
      y: data.y / scale + hy
    };
  };

  return fn;
}